# -*- coding: utf-8 -*-

import os
import csv
import shutil
import cStringIO
import subprocess
import traceback
import logging
import time
import re
import datetime
import operator
from collections import defaultdict
from contextlib import contextmanager
import univention.admin.modules
import univention.admin.uldap
import univention.testing.ucr
import univention.testing.udm as testing_udm
import univention.testing.strings as uts
import univention.testing.utils as utils
from univention.ox.listener_tools import (
	get_credentials, get_master_credentials, ox_call, ox_list_context, ox_list_group, ox_list_resource, ox_list_user,
	get_credentials_for_context, default_context
)
from univention.ox.wait_for_ox_server import wait_for_ox_server
from univention.ox.http.client import Client as OXHttpApi, LoginError
from univention.ox.backend_base import get_ox_integration_class


class OxAccessError(Exception):
	pass


class ExternalCommandValueError(Exception):
	pass


class ExternalCommandExecutionError(Exception):
	pass


logger = None
ox_access_list = ['none', 'webmail', 'pim', 'groupware', 'premium']

CSV_USER_ROW_USERNAME = 'Name'
CSV_USER_ROW_UID = 'Id'
CSV_USER_ROW_GROUPNAME = 'name'
CSV_USER_ROW_MEMBERS = 'members'
CSV_GROUP_ROW_ID = 'id'


def fail(msg, raise_exception=True):
	if raise_exception:
		utils.fail(msg)
	else:
		print(msg)
		return False


def get_access_combination_name(udm_access_name, with_usm):
	return 'oxseforucs_%s%s%s' % (
		udm_access_name,
		'_plus' if udm_access_name == 'pim' else '',  # access definition is called "oxseforucs_pim_plus" but LDAP value is called "pim"
		'_usm' if with_usm else ''
	)


def get_logger():
	global logger
	if not logger:
		logger = logging.getLogger()
		logger.process = logger.info
		logger.setLevel(logging.DEBUG)
		if 'ucs-test' not in [h.get_name() for h in logger.handlers]:
			handler = logging.StreamHandler()
			handler.set_name('ucs-test')
			handler.setFormatter(logging.Formatter(
				fmt='%(asctime)s %(levelname)s: %(message)s',
				datefmt='%Y-%m-%d %H:%M:%S'))
			handler.setLevel(logging.DEBUG)
			logger.addHandler(handler)
	return logger


def convert_ox_access_definitions(flags):
	mapping = {
		'access-active-sync': 'activesync',
		'access-calendar': 'calendar',
		'access-collect-email-addresses': 'collectemailaddresses',
		'access-contacts': 'contacts',
		'access-delegate-tasks': 'delegatetask',
		'access-edit-group': 'editgroup',
		'access-edit-password': 'editpassword',
		'access-edit-public-folder': 'editpublicfolders',
		'access-edit-resource': 'editresource',
		'access-ical': 'ical',
		'access-infostore': 'infostore',
		'access-multiple-mail-accounts': 'multiplemailaccounts',
		'access-olox20': 'olox20',
		'access-publication': 'publication',
		'access-read-create-shared-Folders': 'readcreatesharedfolders',
		'access-subscription': 'subscription',
		'access-syncml': 'syncml',
		'access-tasks': 'tasks',
		'access-usm': 'usm',
		'access-vcard': 'vcard',
		'access-webdav': 'webdav',
		'access-webdav-xml': 'webdavxml',
		'access-webmail': 'webmail',
	}
	result = {}
	flag_set = set(flags.split(','))
	for access_name, flag_name in mapping.items():
		if flag_name in flag_set:
			result[access_name] = 'true'
			flag_set.remove(flag_name)  # remove known flags
		else:
			result[access_name] = 'false'
	if flag_set:
		raise ValueError('Unknown module access definitions found: %r' % (flag_set,))
	return result

ox_access_permission_lists = {
	'oxseforucs_premium': convert_ox_access_definitions('calendar,contacts,delegatetask,editpublicfolders,ical,infostore,readcreatesharedfolders,tasks,vcard,webdav,webdavxml,webmail,collectemailaddresses,multiplemailaccounts,subscription,publication,usm,olox20'),
	'oxseforucs_groupware': convert_ox_access_definitions('calendar,contacts,delegatetask,editpublicfolders,infostore,readcreatesharedfolders,tasks,webmail,collectemailaddresses,multiplemailaccounts,subscription,publication'),
	'oxseforucs_pim_plus': convert_ox_access_definitions('calendar,contacts,delegatetask,tasks,webmail,collectemailaddresses,multiplemailaccounts,subscription,publication'),
	'oxseforucs_webmail': convert_ox_access_definitions('contacts,webmail,collectemailaddresses,multiplemailaccounts,subscription,publication'),

	'oxseforucs_premium_usm': convert_ox_access_definitions('calendar,contacts,delegatetask,editpublicfolders,ical,infostore,readcreatesharedfolders,tasks,vcard,webdav,webdavxml,webmail,collectemailaddresses,multiplemailaccounts,subscription,publication,activesync,usm,olox20'),
	'oxseforucs_groupware_usm': convert_ox_access_definitions('calendar,contacts,delegatetask,editpublicfolders,infostore,readcreatesharedfolders,tasks,webmail,collectemailaddresses,multiplemailaccounts,subscription,publication,activesync,usm'),
	'oxseforucs_pim_plus_usm': convert_ox_access_definitions('calendar,contacts,delegatetask,tasks,webmail,collectemailaddresses,multiplemailaccounts,subscription,publication,activesync,usm'),
	'oxseforucs_webmail_usm': convert_ox_access_definitions('contacts,webmail,collectemailaddresses,multiplemailaccounts,subscription,publication,activesync,usm'),
}


def run_ext_cmd(cmdline):
	print('run_ext_cmd(cmdline={!r})'.format(cmdline))
	# must connect stdin or systemctl will be unhappy (Failed to open /dev/tty)
	cmd_proc = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE, shell=False)
	cmd_out, cmd_err = cmd_proc.communicate()
	cmd_exit = cmd_proc.wait()
	if cmd_exit != 0:
		raise ExternalCommandExecutionError('Command returned an error: exit code={!r}, stdout={!r} stderr={!r}'.format(
			cmd_exit, cmd_out, cmd_err))
	if cmd_err:
		print('stderr={}'.format(cmd_err))
	return cmd_out.strip()


def listuser_csv(username_pattern, contextid):
	print('listuser_csv(username_pattern={!r}, contextid={!r})'.format(username_pattern, contextid))
	cmdline = [
		'/opt/open-xchange/sbin/listuser',
		'--adminuser', 'oxadmin',
		'--adminpass', open('/etc/ox-secrets/context{}.secret'.format(contextid), 'rb').read().strip(),
		'--contextid', str(contextid),
		'--searchpattern', username_pattern,
		'--csv',
	]
	csv_str = run_ext_cmd(cmdline)
	csv_reader = csv.DictReader(cStringIO.StringIO(csv_str))
	for row in csv_reader:
		yield row


def verify_ox_mailenabled_flag(username, contextid, expected_status):
	user_csv = list(listuser_csv(username, contextid))
	if len(user_csv) != 1:
		raise ExternalCommandValueError('Got more than 1 result from listuser_csv({!r}, {!r}).'.format(username, contextid))
	flag_mailenabled = user_csv[0].get('Mailenabled', 'undefined')
	if flag_mailenabled.lower() != str(expected_status).lower():
		raise OxAccessError('Expected flag "mailenabled" in status %r but found status %r.' % (expected_status, flag_mailenabled))


def verify_ox_access_permissions(username, combination_name, contextid):
	print('verify_ox_access_permissions(username={!r}, combination_name={!r}, contextid={!r})'.format(
		username, combination_name, contextid))
	user_csv = list(listuser_csv(username, contextid))
	if len(user_csv) != 1:
		raise ExternalCommandValueError('Got more than 1 result from listuser_csv({}, {}).'.format(username, contextid))
	user_csv = user_csv[0]
	for perm, value in ox_access_permission_lists[combination_name].items():
		print 'Permission: %s=%s  (expected=%s)' % (perm, user_csv[perm], value)
		if user_csv[perm] != value:
			raise OxAccessError('Expected "{}" to be "{}" for "{}", but found "{}". listuser_csv() result was:\n{}\n'.format(
				perm, value, combination_name, user_csv[perm], user_csv))


@contextmanager
def restart_listener():
	"""
	Context manager that will restart the univention-directory-listener when
	exiting.
	"""
	yield
	run_ext_cmd(["service", "univention-directory-listener", "restart"])


class OXTestCredentials(object):
	def __init__(self, username, password):
		self.username = username
		self.password = password


class OXTestUser(object):
	def __init__(self, username, context_id, dn=None, password=None):
		self.username = username
		self.context_id = context_id
		self.password = password
		self.dn = dn


class OXTestEmptyQueue(object):
	_ucr = univention.testing.ucr.UCSTestConfigRegistry()
	_ucr.load()
	listener_queue_base_dir = '/var/spool/univention-directory-listener'

	def __init__(self):
		self._listener_running = True

	def __enter__(self):
		for root, dirs, files in os.walk(self.listener_queue_base_dir):
			current_dir = os.path.basename(root)
			if not current_dir.startswith('ox-'):
				continue
			queue_file_names = [f for f in files if f != 'lock.tmp']
			if queue_file_names:
				print('*** Removing ox listener queue files:\n{}'.format(
					'\n'.join([os.path.join(root, fn) for fn in queue_file_names])))
				self.stop_listener()
				shutil.rmtree(root, ignore_errors=True)
		self.start_listener()
		return self

	def __exit__(self, exc_type, exc_value, etraceback):
		pass

	def start_listener(self):
		if not self._listener_running:
			try:
				utils.wait_for_replication()
			except utils.LDAPReplicationFailed:
				print('Ignoring LDAPReplicationFailed.')
			run_ext_cmd(['/usr/bin/deb-systemd-invoke', 'start', 'univention-directory-listener.service'])
			self._listener_running = True

	def stop_listener(self):
		if self._listener_running:
			try:
				utils.wait_for_replication()
			except utils.LDAPReplicationFailed:
				print('Ignoring LDAPReplicationFailed.')
			run_ext_cmd(['/usr/bin/deb-systemd-invoke', 'stop', 'univention-directory-listener.service'])
			self._listener_running = False


class OXTestEnv(object):
	_ucr = univention.testing.ucr.UCSTestConfigRegistry()
	_ucr.load()

	def __init__(self):
		self._context_ids = set()
		self._users = set()
		self.ox_master_admin, self.ox_master_admin_password = get_master_credentials()
		self._context_admins = dict()
		wait_for_ox_server()

	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_value, etraceback):
		if exc_type:
			print '*** Cleanup OX after exception: %s %s' % (exc_type, exc_value)
		try:
			self.cleanup()
		except:
			print ''.join(traceback.format_exception(exc_type, exc_value, etraceback))
			raise

	def add_user(self, username, context_id=10, dn=None, password=None):  # type: (str, int, Optional[str], Optional[str]) -> None
		self._users.add(OXTestUser(username, context_id, dn, password))

	def add_context(self, context_id):  # type: (int) -> None
		self._context_ids.add(context_id)

	def _create_context(self, context_id, password=None):  # type: (Union[int, str], Optional[str]) -> None
		print('*** Creating OX context {}...'.format(context_id))
		self.add_context(int(context_id))
		if not password:
			password = get_credentials('/etc/ox-secrets/context10.secret')
		cmd = [
			'/usr/sbin/ox-create-context',
			'--context',
			str(context_id),
			'--password',
			password,
		]
		run_ext_cmd(cmd)

	def create_context(self, context_id, password=None):  # type: (Union[int, str], Optional[str]) -> None
		"""
		Creates a new context with the given context id. The password of the default context 10
		is reused if not specified.
		"""
		self._create_context(context_id, password)
		utils.wait_for_replication()
		wait_for_ox_server()
		time.sleep(10)  # race condition: connection works, while the server is shutting down :/
		wait_for_ox_server()

	def create_contexts(self, amount):  # type: (int) -> List[int]
		for _i in xrange(amount):
			print('*** Creating context {}/{}...'.format(_i + 1, amount))
			num = uts.random_int(1000000, 2000000)
			while num in self._context_ids:
				num = uts.random_int(1000000, 2000000)
			self._create_context(num)
		utils.wait_for_replication()
		wait_for_ox_server()
		time.sleep(10)  # race condition: connection works, while the server is shutting down :/
		wait_for_ox_server()
		return sorted(self._context_ids)

	def create_resource_via_udm(self, udm, name=None):  # type: (Union[int, str], Optional[str]) -> None
		name = name or uts.random_username()
		print('*** Creating OX resource {!r} in default context ({!r})...'.format(name, default_context))
		mail_domain = self._ucr.get('mail/hosteddomains').split()[-1].strip()
		dn = udm.create_object(
			'oxresources/oxresources',
			position='cn=oxresources,cn=open-xchange,{}'.format(self._ucr['ldap/base']),
			name=name,
			displayname='{}'.format(uts.random_username()),
			resourceMailAddress='{}@{}'.format(uts.random_username(), mail_domain),
			resourceadmin='$(getent passwd "$username" | cut -d: -f3)'
		)
		return dn

	def create_users(self, udm, amount, context_ids, wait_for_replication=True):
		# type: (object, int, List[Union[int, str]], Optional[bool]) -> List[OXTestUser]
		num_contexts = len(context_ids)
		mail_domain = self._ucr.get('mail/hosteddomains').split()[-1].strip()
		for user_num in xrange(amount):
			context_id = context_ids[user_num % num_contexts]
			print('*** Creating user {}/{} in OX context {}...'.format(user_num + 1, amount, context_id))
			user_password = uts.random_string()
			user_dn, username = udm.create_user(
				mailPrimaryAddress='{}@{}'.format(uts.random_username(), mail_domain),
				password=user_password,
				isOxUser='OK',
				oxContext=context_id,
				wait_for_replication=False,
				check_for_drs_replication=False)
			self.add_user(username, context_id, user_dn, user_password)
		if wait_for_replication:
			utils.wait_for_replication()
		return list(self._users)

	@staticmethod
	def check_if_user_attributes(username, context_id, attribs, raise_exception=True):
		# type: (str, Union[int, str], Dict[str, str], Optional[bool]) -> bool
		"""
		Returns False or raises Exception the if the users attributes do not match those in OX DB.
		Otherwise, True is returned.
		"""
		print('*** Checking attributes of user {!r} in context {!r} via SOAP...'.format(username, context_id))
		import univention.ox.soap.backend  # noqa
		context_id = int(context_id)
		wait_for_ox_server()
		User = get_ox_integration_class('SOAP', 'User')
		user = User.from_ox(context_id, name=username)
		success = True
		for k, v in attribs.items():
			if v == getattr(user, k):
				print('OK: {!r}'.format(k))
			else:
				print('ERROR: {!r} - expected: {!r} found: {!r}'.format(k, v, getattr(user, k)))
				success = False
		if raise_exception and not success:
			utils.fail('Errors checking attributes of user {!r}.')
		return success


	@staticmethod
	def check_csv(reader, expected, unique_id_column):
		# type: (Union[csv.DictReader, List[Dict[str, str]]], Dict[str, Dict[str]], str) -> None
		"""
		Check if all of the dicts in the list expected are in one of the rows
		found in reader.

		:param reader: DictReader/list instance
		:param expected: dict of dicts: unique_id -> expected_csv_line
		:param unique_id_column: str: column name which can be used to uniquely identify a line in the CSV, usually 'id'
		:return: None
		"""
		csv_id_to_row_mapping = dict()
		for row in reader:
			csv_id_to_row_mapping[row[unique_id_column]] = row

		for exp_id, exp_line in expected.items():
			try:
				csv_line = csv_id_to_row_mapping[exp_id]
			except KeyError as exc:
				utils.fail('Could not find expected line with ID "{}" in CSV: {}\nError: {}'.format(exp_id, exp_line, exc))

			for k, v in exp_line.items():
				if not re.match(v, csv_line[k]):
					utils.fail('With ID "{}" in column "{}": expected={} found={}\nin CSV line: {}\nin expected line: {}'.format(
						exp_id, k, repr(v), repr(csv_line[k]), csv_line, exp_line))

	def check_if_context_exists_in_ox_via_csv(self, context_id, raise_exception=True):  # type: (Union[int, str], Optional[bool]) -> bool
		"""
		Returns False or raises Exception the if requested context does not exist in OX DB.
		Otherwise, True is returned.
		"""
		print('*** Check if context {} exists in OX via CSV...'.format(context_id))
		wait_for_ox_server()
		context_csv = None
		for tries in xrange(5, 0, -1):
			try:
				context_csv = ox_list_context(self.ox_master_admin, self.ox_master_admin_password)
				break
			except ExternalCommandExecutionError:
				print('Cannot read contexts... waiting {} times...'.format(tries - 1))
				time.sleep(5)
				continue

		if not context_csv:
			if raise_exception:
				utils.fail('Could not read contexts / no contexts found')
			else:
				return False

		for row in context_csv:
			if row['id'] == str(context_id):
				print('*** OK: found context {} in OX.'.format(context_id))
				return True
		if raise_exception:
			utils.fail('Could not find context {!r} in output of "listcontext".'.format(context_id))
		return False

	@staticmethod
	def check_if_context_exists_in_ox_via_soap(context_id, raise_exception=True):
		# type: (Union[int, str], Optional[bool]) -> bool
		"""
		Returns False or raises Exception the if requested context does not exist in OX DB.
		Otherwise, True is returned.
		"""
		print('*** Checking if context {!r} exists in OX via SOAP...'.format(context_id))
		import univention.ox.soap.backend  # noqa
		context_id = int(context_id)
		wait_for_ox_server()
		Context = get_ox_integration_class('SOAP', 'Context')
		contexts = Context.list(0, context_id)
		if len(contexts) > 1:
			print('WARNING: found more than 1 context when searching for {!r}: {!r}'.format(context_id, contexts))
		res = context_id in [c.id for c in contexts]
		if raise_exception and not res:
			utils.fail('Could not find context {!r} in SOAP reply.'.format(context_id))
		return res

	@classmethod
	def check_if_context_exists_in_ldap(cls, context_id, raise_exception=True):
		# type: (Union[int, str], Optional[bool]) -> bool
		print('*** Check if context exists in LDAP ...')
		context_objs = cls.get_udm_objects('oxmail/oxcontext')

		print('Available context IDs in LDAP: {}'.format(
			','.join([context_obj.get('contextid') for context_obj in context_objs])))
		for context_obj in context_objs:
			if context_obj['name'] == 'context{}'.format(context_id):
				assert (context_obj['contextid'] == str(context_id))
				print('*** OK: found context {} in LDAP.'.format(context_id))
				return True
		msg = 'Could not find context id {!r} in LDAP.'.format(context_id)
		if raise_exception:
			utils.fail(msg)
		else:
			print('*** {}'.format(msg))
			return False

	def check_if_user_exists_in_context(self, username, context_id, expected=True, raise_exception=True):
		# type: (str, Union[int, str], Optional[bool], Optional[bool]) -> bool
		"""
		Checks if a user exists within a certain context.
		Returns False/Raises an exception if the expected result differs from actual result.
		"""
		context_admin_username, context_admin_password = get_credentials_for_context(context_id)
		found = False
		print('*** Looking for OX user "{}" in OX context {} ...'.format(username, context_id))
		for row in ox_list_user(
			context_admin_username,
			context_admin_password,
			context_id,
			searchpattern=username
		):
			print('row={}'.format(row[CSV_USER_ROW_USERNAME]))
			if row[CSV_USER_ROW_USERNAME] == username:
				print('*** OK: found user "{}" in context {}.'.format(username, context_id))
				found = True
				break
		if found != expected:
			msg = 'Unexpected result while testing if user "{}" exists in context {}: expected={}  actual_result={}'.format(username, context_id, expected, found)
			if raise_exception:
				utils.fail(msg)
			else:
				print('*** {}'.format(msg))
				return False
		print('*** OK: Expected result while testing if user "{}" exists in context {}: expected={}  actual_result={}'.format(username, context_id, expected, found))
		return True

	def check_if_users_exists_in_context(self, users, context_id):  # type: (List[OXTestUser], Union[int, str]) -> None
		context_admin_username, context_admin_password = get_credentials_for_context(context_id)
		csv_username_to_row = dict()
		for row in ox_list_user(context_admin_username, context_admin_password, context_id):
			csv_username_to_row[row[CSV_USER_ROW_USERNAME]] = row
		for user in users:
			if user.username not in csv_username_to_row:
				utils.fail('User "{}" was not found in context {}.'.format(user.username, context_id))
		print('*** Found all {} users in context {}.'.format(len(users), context_id))

	@staticmethod
	def check_if_resource_exists_via_csv(context_id, name, raise_exception=True):
		# type: (Union[int, str], str, Optional[bool]) -> bool
		"""
		Returns False or raises Exception the if requested resource does not exist in OX DB.
		Otherwise, True is returned.
		"""
		print('*** Checking if resource {!r} exists in context {!r} via CSV...'.format(name, context_id))
		user, password = get_credentials_for_context(context_id)
		found = False
		for row in ox_list_resource(user, password, context_id, searchpattern=name):
			if row['name'] == name:
				found = True
		if not found and raise_exception:
			utils.fail('Resource {!r} was not found in context {!r}.'.format(name, context_id))
		return found

	@staticmethod
	def check_if_resource_exists_via_soap(context_id, name, raise_exception=True):
		# type: (Union[int, str], str, Optional[bool]) -> bool
		"""
		Returns False or raises Exception the if requested resource does not exist in OX DB.
		Otherwise, True is returned.
		"""
		print('*** Checking if resource {!r} exists in context {!r} via SOAP...'.format(name, context_id))
		import univention.ox.soap.backend  # noqa
		context_id = int(context_id)
		wait_for_ox_server()
		Resource = get_ox_integration_class('SOAP', 'Resource')
		resources = Resource.list(context_id, name)
		if len(resources) > 1:
			print('WARNING: found more than 1 resource when searching for {!r}: {!r}'.format(name, resources))
		res = name in [r.name for r in resources]
		if raise_exception and not res:
			utils.fail('Could not find resource {!r} in SOAP reply.'.format(name))
		return res

	def get_ox_id_to_name_mapping(self, context_id):  # type: (str) -> Dict[str, str]
		context_admin_username, context_admin_password = get_credentials_for_context(context_id)
		res = dict()
		for row in ox_list_user(context_admin_username, context_admin_password, context_id):
			row_username = row[CSV_USER_ROW_USERNAME]
			row_user_id = row[CSV_USER_ROW_UID]
			res[row_user_id] = row_username
		return res

	def get_ox_group(self, group_name, context_id, raise_exception=True):  # type: (str, str, Optional[bool]) -> List[str]
		print('*** Looking for group "{}" in context {}.'.format(group_name, context_id))

		context_admin_username, context_admin_password = get_credentials_for_context(context_id)
		for row in ox_list_group(context_admin_username, context_admin_password, context_id):
			if group_name == row[CSV_USER_ROW_GROUPNAME]:
					print('*** Found group "{}" in context {}.'.format(group_name, context_id))
					return row
		fail('Group "{}" was not found in context {}.'.format(group_name, context_id), raise_exception)

	def check_if_group_exists_in_context(self, group_name, context_id, members=None, raise_exception=True):  # type: (str, str, Optional[List[str]], Optional[bool]) -> bool
		members = members or []
		print('*** Looking for group "{}" in context {}{}.'.format(
			group_name,
			context_id,
			' with members {}'.format(members) if members else ''))

		import univention.ox.soap.backend  # noqa
		context_id = int(context_id)
		wait_for_ox_server()
		Group = get_ox_integration_class('SOAP', 'Group')
		User = get_ox_integration_class('SOAP', 'User')
		groups = Group.list(context_id, group_name)
		if len(groups) > 1:
			print('WARNING: found more than 1 group when searching for {!r}: {!r}'.format(context_id, groups))
		found = False
		for group in groups:
			if group.name == group_name:
				print('*** Found group {!r} in context {!r}.'.format(group_name, context_id))
				found = True
				for member in members:
					ox_id = User.from_ox(context_id, name=member).id
					if ox_id in group.members:
						print('*** Found member {!r}.'.format(member))
					else:
						fail('Found group {!r} in context {!r} with members {!r}. Member {!r} is missing.'.format(group_name, context_id, group.members, member), raise_exception)
		if not found:
			fail('Group {!r} was not found in context {!r}.'.format(group_name, context_id), raise_exception)

	@staticmethod
	def check_if_user_can_login(username, password, expected=True, raise_exception=True):  # type: (str, str, Optional[bool], Optional[bool]) -> bool
		"""
		Checks if a user is able to login via HTTP API.
		Returns False/Raises an exception if the expected result differs from actual result.
		"""
		can_login = True
		wait_for_ox_server()
		try:
			client = OXHttpApi(username, password)
		except LoginError:
			can_login = False

		if can_login != expected:
			msg = 'Unexpected result while testing HTTP API login for user "{}": expected={}  actual_result={}'.format(username, expected, can_login)
			if raise_exception:
				utils.fail(msg)
			else:
				print('*** {}'.format(msg))
				return False
		print('*** OK: Expected result while testing HTTP API login for user "{}": expected={}  actual_result={}'.format(username, expected, can_login))

		if expected:
			ox_user = client.get_user()
		return True

	def delete_user(self, username, context_id=10):  # type: (str, Union[int, str]) -> None
		oxadmin_username, oxadmin_password = get_credentials_for_context(int(context_id))
		deleteuser = [
			'deleteuser',
			'--username={}'.format(username),
			'--adminuser={}'.format(oxadmin_username),
			'--adminpass={}'.format(oxadmin_password),
			'--contextid={}'.format(context_id),
		]
		ox_call(deleteuser, 'com.openexchange.admin.rmi.exceptions.NoSuchObjectException', get_logger())

	def delete_context(self, context_id, wait_for_replication=True):  # type: (int, Optional[bool]) -> None
		cmd = [
			'/usr/sbin/univention-directory-manager',
			'oxmail/oxcontext',
			'remove',
			'--dn=cn=context{},cn=open-xchange,{}'.format(context_id, self._ucr.get('ldap/base')),
		]
		run_ext_cmd(cmd)
		if wait_for_replication:
			utils.wait_for_replication()

		cmd = [
			'deletecontext',
			'--contextid={}'.format(context_id),
			'--adminuser={}'.format(self.ox_master_admin),
			'--adminpass={}'.format(self.ox_master_admin_password),
		]
		ox_call(cmd, 'com.openexchange.admin.rmi.exceptions.NoSuchObjectException', get_logger())

	def cleanup(self):  # type: () -> None
		""" Cleanup all objects created by the OX test environment """
		wait_for_ox_server()
		print('Performing OXTestEnv cleanup...')
		for user in self._users:
			self.delete_user(user.username, user.context_id)
		try:
			utils.wait_for_replication()
		except utils.LDAPReplicationFailed:
			pass
		for context_id in self._context_ids:
			self.delete_context(context_id, False)
		self.long_wait_for_replication(False)
		print('OXTestEnv cleanup done')

	@staticmethod
	def get_udm_objects(udm_module_name):
		lo, position = univention.admin.uldap.getMachineConnection()
		univention.admin.modules.update()
		module_context = univention.admin.modules.get(udm_module_name)
		univention.admin.modules.init(lo, position, module_context)
		return module_context.lookup({}, lo, 'name=*')

	def get_user_count(self, context_ids):
		num = 0
		for context_id in context_ids:
			context_admin_username, context_admin_password = get_credentials_for_context(context_id)
			num += len(['x' for row in ox_list_user(context_admin_username, context_admin_password, context_id) if row[CSV_USER_ROW_USERNAME] != context_admin_username])
		return num

	def long_wait_for_replication(self, user_count_grows):
		while True:
			try:
				utils.wait_for_replication()
				break
			except utils.LDAPReplicationFailed:
				print('LDAPReplicationFailed, checking if listener is still {} users...'.format(
					'creating' if user_count_grows else 'deleting'))
				user_count = self.get_user_count(self._context_ids)
				print('User count in all used contexts is {}... sleeping 20s...'.format(user_count))
				time.sleep(20)
				user_count2 = self.get_user_count(self._context_ids)
				print('User count in all used contexts is {} now.'.format(user_count2))
				if user_count_grows:
					if user_count2 > user_count:
						print('User count is still growing... waiting again.')
					else:
						print('User count not growing - continuing...')
						break
				else:
					if user_count2 < user_count:
						print('User count is still falling... waiting again.')
					else:
						print('User count not falling - continuing...')
						break

	@staticmethod
	def restart_ox_server():
		print('Restarting OX server...')
		print(run_ext_cmd(['deb-systemd-invoke', 'restart', 'open-xchange.service']))
		wait_for_ox_server()

	def user_creation_performance_test(self, num_contexts, num_users):
		print('*** Creating {} users in {} contexts...'.format(num_users, num_contexts))
		with Timing() as timer, testing_udm.UCSTestUDM() as udm:
			timer.start('context creation')
			context_ids = self.create_contexts(num_contexts)
			print(timer.stop('context creation', True))
			print('*** Created contexts: {}'.format(context_ids))
			if len(context_ids) != num_contexts:
				utils.fail('Wanted to create {} contexts but {} were created.'.format(num_contexts, len(context_ids)))

			timer.start('user creation (w/o waiting for listener)')
			users = self.create_users(udm, num_users, context_ids, wait_for_replication=False)
			print(timer.stop('user creation (w/o waiting for listener)', True))
			if len(users) != num_users:
				utils.fail('Wanted to create {} users but {} were created.'.format(num_users, len(users)))

			timer.start('wait_for_replication()')
			self.long_wait_for_replication(True)
			print(timer.stop('wait_for_replication()', True))

			context_to_users = defaultdict(list)
			for user in users:
				context_to_users[user.context_id].append(user)
			print('Context/User distribution:')
			for k in sorted([int(k) for k in context_to_users.keys()]):
				print('{}: {}'.format(k, len(context_to_users[k])))

			print('*** Verifying users...')
			timer.start('user verify')
			for context_id, users in context_to_users.items():
				for user in users:
					utils.verify_ldap_object(user.dn, {'oxAccess': ['premium'], 'oxContextIDNum': [str(user.context_id)]})
				self.check_if_users_exists_in_context(users, context_id)
			print(timer.stop('user verify', True))
			return timer


class Timing(object):
	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_value, etraceback):
		print('Timing:\n{}\n'.format(self))

	def __init__(self):
		self.times = dict()

	def start(self, name):
		self.times[name] = dict(start=datetime.datetime.now())

	def stop(self, name, verbose=False):
		self.times[name]['stop'] = datetime.datetime.now() - self.times[name]['start']
		if verbose:
			return '{}: {}'.format(name, self.human_time(self.times[name]['stop']))

	@property
	def sorted_list(self):
		return sorted([(v['start'], v.get('stop'), k) for k, v in self.times.items()], key=operator.itemgetter(0))

	@staticmethod
	def human_time(tdelta):
		seconds = tdelta.seconds
		h = seconds / 3600
		seconds -= h * 3600
		m = seconds / 60
		seconds -= m * 60
		return '{}h {}min {}sec {}ms'.format(h, m, seconds, tdelta.microseconds / 1000)

	def __str__(self):
		return '\n'.join(
			['{}: {}'.format(t[2], self.human_time(t[1]) if t[1] else 'no end') for t in self.sorted_list])
